-- tools for tuning and measuring design performance
--
-- when a train departs a stop, stats about its stay are printed.
--
-- when a train departs a `==>`-marked stop, inventory in
-- circuit-attached storage is adjusted to bring the signal counts on
-- the circuit to zero (after adjusting for read-from-train if set).
--
-- adds a 'bp' command to reset circuit-attached storage for selection
--
-- adds a 'iod' command, "instead of deconstructing" an item calls a
-- function with `e` set to the entity and `p` set to the player, that
-- cancels deconstruction and runs the command parameter instead, e.g.
--       /iod e.insert{name='iron-plate',count=1e9}
-- and then wave a decon planner over everything you want to stuff with
-- iron plates.
--
-- adds a 'cheat' command to make all blueprint activity instant and
-- give equipment, speed and range bonuses.
--
npg={} for k in next,_G do npg[k]=true end -- non-persisted globals

require 'utils.all'
quyxkh={}
empty = setmetatable({},
	{__newindex = function() error 'assigning to empty' end})
di=defines.inventory

trainstop = {}  -- stops trains are waiting at
	    -- and their cargo on arrival (see train_changed_state)
tickstop = {} -- stops whose names start ==> and end %n are added to
	    -- this automatically, `n` being the reset-attached-storage
	    -- beat when no train is stopped there. You can also
	    -- add any entity manually, `/bp set_tickstop(e,300)
--
local tsrec = function(t,ts) -- record of a train's arrival at a stop
    local sr = { tick = game.tick,
	    stop = ts,
	    pos={x=ts.position.x,y=ts.position.y},
	    ekey = ekey(ts),
	    cargo = t.get_contents() or {}
	    }
    for k,v in next,t.get_fluid_contents() or empty do
	sr.cargo[k]=v
	end
    return sr
    end

function reset_network_storage(e,w,readout)
    -- this code's completely unaware of bars and filters
    if not e then return end
    if not readout then
	readout = read_signals(e,w,circuit[e.type] or 1) or empty
	end
    local chests, cslots = {}, 0
    local tanks = {['']={capacity=0}}
    local removed = { item={}, fluid={} }
    local ri,rf = removed.item, removed.fluid
    local added = { item={}, fluid={} }
    local ai,af = added.item, added.fluid

    -- clear all of anything showing excess inventory, any of it that
    -- was wanted will be redistributed next.
    for e in iterate_network(e,w) do
	local b = e.get_inventory(di.chest)
	local bn = b and b.hasbar() and b.getbar()
	if bn then
	    chests[#chests+1] = b
	    cslots = cslots + bn-1
	    for i,v in next,readout.item do if v>0 then
		ri[i] = (ri[i] or 0)
		    + b.remove{name=i,count=1e9}
		end end
	elseif #e.fluidbox > 0 then
	    local b,f = e.fluidbox f = b[1]

	    local t = f and f.name or ''

	    tanks[t] = tanks[t] or {capacity=0}
	    tanks[t][#tanks[t]+1] = b

	    local c = b.get_capacity(1)
	    tanks[t].capacity = tanks[t].capacity + c

	    local r = readout.fluid[t]
	    if r and r>0 then
		rf[t] = (rf[t] or 0) + f.amount
		b[1] = nil
		end
	else
	    -- not storage so ntd
	    end
	end

    -- amend item wants
    for i,c in next,readout.item do
	local r = ri[i] or 0
	if r-c > 0 then
	    local pcs = (r-c+0.001)/cslots -- per-chest-slot
	    local pcx = 0
	    ai[i] = ai[i] or 0
	    for _,b in next,chests do
		local add = pcs*(b.getbar()-1) + pcx
		local a = add>=1 and b.insert{name=i,count=add-add%1} or 0
		if r>0 then ri[i] = ri[i]-a
		       else ai[i] = ai[i]+a
		       end
		--[[ print(string.format('***pcs=%g,pcx=%g,add=%g,a=%g,r-c=%g',
					pcs,pcx,add,a,r-c)) c=c+a ]]
		pcx = add%1
		end
	    end
	end

    -- amend fluid wants
    for f,a in next,readout.fluid do
	local r = rf[f] or 0
	if quyxkh.debug then psl{rf={f=r},r=r,a=a} end
	if r-a > 0 then
	    if not tanks[f] then tanks[f]={capacity=0} end
	    local c = tanks[f].capacity
	    while c < r-a and #tanks[''] > 0 do
		b=table.remove(tanks[''])
		tanks[f][#tanks[f]+1]=b
		c = c + b.get_capacity(1)
		tanks[f].capacity = c
		end
	    local added = 0
	    if quyxkh.debug then
		psl{capacity=c,tanks=#tanks[f]}
		end
	    for t=1,#tanks[f] do
		local x = tanks[f][t][1] or {amount=0, temperature=15}
		local m = tanks[f][t].get_capacity(1)/c * (r-a)
		tanks[f][t][1]={name=f,
		    temperature=x.temperature,
		    amount=x.amount + m}
		added = added + m
		end
	    if r > 0 then rf[f] = (rf[f] or 0)-added
		 else af[f] = (af[f] or 0)+added
		 end
	    end
	end

    -- update in-game benchmark stats
    do local stat=game.forces.benchmark.item_production_statistics
	for k,v in next,removed.item do stat.on_flow(k,-v) end
	for k,v in next,added.item   do stat.on_flow(k,v) end
	end
    do local stat=game.forces.benchmark.fluid_production_statistics
	for k,v in next,removed.fluid do stat.on_flow(k,-v) end
	for k,v in next,added.fluid do stat.on_flow(k,v) end
	end
    end

sname = {} for k,v in next,defines.train_state do sname[v]=k end
print(table.concat(table.pack
    {'act','#','item','tick','wait','tprod','sname','sx','sy','tname','tid','del','pu','keep'}[1],'|'))

on('train_changed_state',function(ev)
    local t,sr	= ev.train sr = trainstop[t.id]
    local ts,srs = t.station, sr and sr.stop
    if sname[t.state] == 'wait_station' then
	-- s/b 'wait_stop'^, trains wait at particular stops
	sr = tsrec(t,ts)
	trainstop[t.id] = sr
	trainstop[sr.ekey] = sr
    elseif sr then -- not wait_station and have sr, so leaving
	local delivered, picked_up, kept, prod = {}, {}, {}, {}
	--
	local newcargo = t.get_contents() or {}
	for k,v in next,t.get_fluid_contents() or empty do
	    newcargo[k]=v
	    end
	--
	for k,v in next,sr.cargo do
	    if (newcargo[k] or 0) < v then
		delivered[k] = v - (newcargo[k] or 0)
		end
	    end
	for k,v in next,newcargo do
	    if (sr.cargo[k] or 0) < v then
		picked_up[k] = v - (sr.cargo[k] or 0)
		end
	    if (sr.cargo[k] or 0) > 0 then
		kept[k] = v>sr.cargo[k] and sr.cargo[k] or v
		end
	    end
	--
	local tn = t.locomotives and t.locomotives.front_movers
	tn = tn and tn[1].backer_name or '(?)'
	local sn = sr.stop.valid and sr.stop.backer_name or '(stop gone)'
	local recordpu = false
	local recorddel = false
	if sn:sub(1,3)=='==>' then
	    recordpu  = settings.global['benchmark-pu-is-prod'].value
	    recorddel = settings.global['benchmark-del-is-cons'].value
	    end
	local iprod = game.forces.player.item_production_statistics
	local fprod = game.forces.player.fluid_production_statistics
	--
	local lines, dur = selfmeta{__call=csv,csv='|'}, ev.tick-sr.tick
	for k,v in next,delivered do
	    local stats = game.item_prototypes[k] and iprod or fprod
	    if recorddel then stats.on_flow(k,-v) end
	    local prod = stats.get_input_count(k)
	    lines(':del',v,k,ev.tick,dur,prod,sn,sr.pos.x,sr.pos.y,tn,t.id,v,'','')
	    end
	for k,v in next,picked_up do
	    local stats = game.item_prototypes[k] and iprod or fprod
	    if recordpu then stats.on_flow(k,v) end
	    local prod = stats.get_input_count(k)
	    lines(':pu', v,k,ev.tick,dur,prod,sn,sr.pos.x,sr.pos.y,tn,t.id,'',v,'')
	    end
	for k,v in next,kept do
	    lines(':kept',v,k,ev.tick,dur,'',sn,sr.pos.x,sr.pos.y,tn,t.id,'','',v)
	    end
	if #lines == 0 then
	    lines(':zombie','','',ev.tick,dur,'',sn,sr.pos.x,sr.pos.y,tn,t.id,'','','')
	    end
	lines=table.concat(lines,'\n\t')
	print(lines)
	--
	if srs.valid and srs.backer_name:sub(1,3)=='==>' then
	    local rft = (srs.get_control_behavior() or empty).read_from_train
	    --
	    sigs = read_signals(srs,'red')
	    if rft then
		for k,v in next,newcargo do
		    local it=game.item_prototypes[k] and 'item' or 'fluid'
		    sigs[it][k] = (sigs[it][k] or 0) - v
		    end
		end
	    reset_network_storage(sr.stop,'red',sigs)
	    --
	    sigs = read_signals(srs,'green')
	    if rft then
		for k,v in next,newcargo do
		    local it=game.item_prototypes[k] and 'item' or 'fluid'
		    sigs[it][k] = (sigs[it][k] or 0) - v
		    end
		end
	    reset_network_storage(sr.stop,'green',sigs)
	    --
	    end
	--
	trainstop[t.id] = nil
	trainstop[sr.ekey] = nil
	end
    end)

on('train_created',function(ev)
    -- gut check says this logic's unreliable...
    local t,ts,tsk = ev.train ts=t.station if ts then tsk=ekey(ts) end
    if ts and ev.old_train_id_1 and trainstop[ev.old_train_id_1] then
	trainstop[ev.train.id]=trainstop[ev.old_train_id_1]
	trainstop[ev.old_train_id_1]=nil
	end
    if ts and ev.old_train_id_2 and trainstop[ev.old_train_id_2] then
	trainstop[ev.train.id]=trainstop[ev.old_train_id_2]
	trainstop[ev.old_train_id_2]=nil
	end
    end)

function check_tickstop(ev)
    -- tickstop[interval] = {stop1,stop2,stop3,…}
    for interval,stops in next,tickstop do
	if ev.tick % interval == 0 then
	    for n,ts in next,stops do
		while ts and not ts.valid do
		    stops[n]=stops[#stops]
		    stops[#stops]=nil
		    ts=stops[n]
		    end
		if ts and not trainstop[ekey(ts)] then
		    -- no train here, it's time to reset anyway
		    reset_network_storage(ts,'red')
		    reset_network_storage(ts,'green')
		    end
		end
	    if #stops == 0 then
		tickstop[interval] = nil
		if not next(tickstop) then
		    on('tick',nil)
		    end
		end
	    end
	end
    end
--
function set_tickstop(e,n)
    for interval,stops in next,tickstop do
	for n,stop in next,stops do
	    while stop and (not stop.valid or
		    e.position.x == stop.position.x
		    and e.position.y == stop.position.y) do
		stops[n]=stops[#stops]
		stops[#stops]=nil
		stop=stops[n]
		end
	    if #stops == 0 then
		tickstop[interval] = nil
		if not next(tickstop) then
		    on('tick',nil)
		    end
		end
	    end
	end
    if n then
	if not tickstop[n] then
	    if not next(tickstop) then
		on('tick',check_tickstop)
		end
	    tickstop[n] = {}
	    end
	tickstop[n][#tickstop[n]+1]=e
	end
    end

local function maybe_update_tickstop(ts,old_name) -- train-stop renamed
    -- this next could be more efficient but it's for an unusual user
    -- action and very few specially-marked stops so I don't care
    for interval,stops in next,tickstop do
	for n,stop in next,stops do
	    while stop and (not stop.valid or
		    ts.position.x == stop.position.x
		    and ts.position.y == stop.position.y) do
		stops[n]=stops[#stops]
		stops[#stops]=nil
		stop=stops[n]
		end
	    if #stops == 0 then
		tickstop[interval] = nil
		if not next(tickstop) then
		    on('tick',nil)
		    end
		end
	    end
	end
    if ts.backer_name:sub(1,3) == '==>' then
	local _,_,n = ts.backer_name:find('%%(%d+)$') n=0+(n or 0)
	set_tickstop(ts,n)
	end
    end
on('entity_renamed',function(ev)
    if ev.entity.type == 'train-stop' then
	maybe_update_tickstop(ev.entity,ev.old_name)
	end
    end)

function bp_gpsl(command)
    command.parameter='gpsl('..command.parameter..')'
    return bp_cmd(command)
    end
function bp_cmd(command)
    if type(command)=='string' then
	command={player_index=game.player.index,parameter=command}
	end
    local p = game.players[command.player_index]
    local e = p.selected
    if command.parameter then
	local cmdok, cmd = load(
	    'local bp return function(e,p,b1,b2,b3,b4)'
					..command.parameter..'\nend'
	    , command.parameter)

	if cmdok then
	    cmdok, cmd = pcall(cmdok)
	    end
	if cmdok then
	    game.print(p.name..' (bp): '..command.parameter)
	    cmdok, cmd = pcall(cmd,e,p,B(e,1),B(e,2),B(e,3),B(e,4))
	    end
	if not cmdok then
	    p.print(cmd)
	    end
    elseif e then
	reset_network_storage(e,'red')
	reset_network_storage(e,'green')
	end
    end

-- [[ ** `iod` -- "instead of deconstructing" support ** ]]
-- Deconstruction works in two phases, mark each, then do all.  This
-- hooks both to allow cooperating with other mods' responses.
-- on_marked_for_deconstruction puts each entity in a list, then
-- on_player_deconstructed_area walks the list, and anything still
-- marked for deconstruction has the hook run.  The `insta` helper here
-- destroys it instead of sending a bot to mine it, other mods e.g.
-- screenshotters or list builders that hijack decon-planner selection
-- can `e.cancel_deconstruction()` the entity in their mfd hook.

iodcmd={} -- [pidx]={count=n, string=command.parameter, iod=function}
ioddefault={} -- [pidx]=mfd hook
iodlist=htm{} -- []=remembered marked_for_deconstruction event
iodarea={} -- area passed to pda
--
function clear_iod(player_index)
    iodcmd[player_index or game.player.index] = nil
    if not next(iodcmd) and not next(ioddefault) then
	on('marked_for_deconstruction',nil)
	on('player_deconstructed_area',nil)
	end
    end
--
function do_iod(ev)
    local pidx = ev.player_index
    local p,e,c = game.players[pidx], ev.entity
    e.cancel_deconstruction(p.force)
    if iodcmd[pidx] then
	iodcmd[pidx].count = iodcmd[pidx].count + 1
	c=iodcmd[pidx].iod
    else
	c=ioddefault[pidx]
	end
    local rc,r = xpcall(c,debug.traceback,ev,e,p,B(e,1),B(e,2),B(e,3),B(e,4))
    if rc ~= true then print('**'..r) end
    return rc,r
    end
--
local scratchtile = {name=false,position={}}
function maybe_iod(ev)
    local pidx=ev.player_index
    if not iodcmd[pidx] and not ioddefault[pidx] then return end
    if ioddefault[pidx]==instakill then
	local inhand = (game.players[pidx] or empty).cursor_stack or empty
	if not inhand.valid_for_read then inhand = empty end
	if inhand.is_blueprint or inhand.is_blueprint_book
	    or inhand.name == 'rail' or inhand.type=='copy-paste-tool' then
	    instakill(ev)
	elseif ev.entity.name == 'deconstructible-tile-proxy' then
	    local p = ev.entity.position
	    local t = ev.entity.surface.get_tile(p.x,p.y)
	    if t.hidden_tile then
		scratchtile.name=t.hidden_tile
		scratchtile.position=p
		ev.entity.surface.set_tiles{scratchtile}
		end
	else
	    iodlist:insert(ev)
	    end
    else
	iodlist:insert(ev)
	end
    end
--
function maybe_finish_iod(ev)
    local pidx=ev.player_index
    local p=game.players[pidx]
    iodarea=ev.area
    if not iodcmd[pidx] and not ioddefault[pidx] then
	for i=1,#iodlist do iodlist[i]=nil end
	return end
    for k,ev in iodlist do local e = ev.entity
	if e and e.valid then
	    if e.to_be_deconstructed(p.force)
		then  do_iod(ev)  end
	    iodlist[k]=nil
	    end
	end
    if iodcmd[pidx] then
	game.print(p.name..' (iod) [[ '..  iodcmd[pidx].string..' ]]'
		..' applied to '..iodcmd[pidx].count..' entity/ies')
	clear_iod(pidx)
	end
    end
function set_iod(player_index,t)
    if not next(iodcmd) and not next(ioddefault) then
	on('marked_for_deconstruction',maybe_iod)
	on('player_deconstructed_area',maybe_finish_iod)
	end
    if type(t)=='table' then
        iodcmd[player_index]=t
    elseif type(t)=='function' then
        ioddefault[player_index]=t
    elseif type(t)=='nil' then
	ioddefault[player_index]=nil
	clear_iod(player_index)
        end
    end
--
function iod(command)
    if type(command)~='table'
	then command={parameter=command,player_index=game.player.index}
	end
    local pidx=command.player_index
    local p=game.players[pidx]
    if not command.parameter then
	if iodcmd[pidx] then
	    p.print('iod cancelled')
	    end
	clear_iod(pidx)
	return
	end
    local iodok, iod = load(
	'local iod return function(ev,e,p,b1,b2,b3,b4)'
				    ..command.parameter..'\nend'
	, command.parameter)
    if iodok then
	iodok, iod = pcall(iodok)
	end
    if iodok then
	set_iod(pidx,{count=0,string=command.parameter,iod=iod})
	end
    if not iodok then
	p.print(iod)
	end
    end

function instabuild(ev)
    local e = ev.type and ev or ev.created_entity
    if e.type=='entity-ghost' then
	local inserts = e.item_requests
	if inserts then e.item_requests=nil end
	local _,e2 = e.revive()
	if e2 then
	    for k,v in next,inserts or empty do
		e2.insert{name=k,count=v}
		end
	else
	    e.item_requests = inserts
	    end
	end
    end
--
--[[mfd='marked_for_deconstruction']]
function instakill(ev)
    local e = ev.type and ev or ev.entity
    if not e.valid then return end
    e.destroy()
    end

function bp_insta(command)
    on('built_entity',instabuild)
    set_iod(command.player_index,instakill)
    --
    local p = game.players[command.player_index]
    if p.force.current_research then
	p.force.research_progress=1
	end
    --
    on('research_started',function(ev)
	if ev.research.force.current_research then
	    ev.research.force.research_progress=1
	    end
	end)
    --
    end

function bp_cheat(command)
    local pidx=command.player_index
    local p=game.players[pidx]
    p.cheat_mode=true
    if p.force.current_research then
	p.force.research_progress=1
	end
    do local _=p.force
	-- _.research_all_technologies() does one space level. I want
	-- max pre-space:
	for k,v in pairs(_.technologies) do
	    r= not v.prototype.hidden
	    for _,i in next,v.research_unit_ingredients do
		r = r and i.name~='space-science-pack'
		end
	    v.researched=r
	    end
	_.enable_all_recipes()
	-- _.character_running_speed_modifier=1.25
	_.character_reach_distance_bonus=64
	_.character_build_distance_bonus=64
	_.character_resource_reach_distance_bonus=64
	_.character_inventory_slots_bonus=40
	end
    if not p.character then
	p.create_character() p.clear_items_inside()
	end
    bp_insta(command)
    local count = false
    local seen={}
    for _,v in next,{
	    1000,'rail','train-stop','rail-signal'
	    ,10,'locomotive',30,'cargo-wagon'
	    ,'small-electric-pole','medium-electric-pole','big-electric-pole'
	    ,'substation',1,'electric-energy-interface'
	    ,'arithmetic-combinator','decider-combinator','constant-combinator'
	    ,'red-wire' ,'green-wire'
	    ,'stack-inserter' ,'stack-filter-inserter'
	    ,'express-transport-belt' ,'express-underground-belt'
	    ,'express-splitter' ,'steel-chest'
	    ,'construction-robot'
	    ,'logistic-chest-requester','logistic-chest-active-provider'
	    ,'logistic-chest-passive-provider', 'logistic-chest-storage'
	    ,'roboport'
	    ,1,'power-armor-mk2'
	    ,1,'submachine-gun',1000,'uranium-rounds-magazine'
	    ,1,'combat-shotgun',1000,'piercing-shotgun-shell'
	    ,'rail-chain-signal'
	} do
        if type(v) == 'number' then
            count = v
        elseif not game.item_prototypes[v] then
	    p.print(string.format('no %s prototype?',v))
	    count=false
	else
	    if not seen[v] then p.remove_item{name=v,count=1e9} seen[v]=true end
	    p.insert{name=v, count=count or game.item_prototypes[v].stack_size}
            count=false
            end
        end

    local grid=p.get_inventory(di.character_armor)[1].grid
    for _,v in next,{1,'fusion-reactor-equipment',8,'exoskeleton-equipment'
	    ,4,'battery-mk2-equipment',1,'personal-roboport-mk2-equipment'
	    ,4,'battery-mk2-equipment'
	    --
	    ,12,'energy-shield-mk2-equipment'
	    ,3,'fusion-reactor-equipment'
	    ,3,'personal-roboport-mk2-equipment'
	    ,4,'battery-mk2-equipment'
	  } do
        if type(v) == 'number' then
            count = v
        elseif not game.item_prototypes[v] then
	    p.print(string.format('no %s prototype?',v))
	    count=false
	else
	    if not seen[v] then p.remove_item{name=v,count=1e9} seen[v]=true end
	    if not count then
		count=game.item_prototypes[v].stack_size
		end
	    for _=1,count do
		local e=grid.put{name=v}
		if e then
		    count = count-1
		    if v:find'^battery' then
			e.energy = e.max_energy
			end
		    end
		end
	    if count>0 then p.insert{name=v,count=count} end
	    count=false
	    end
	end
    end

for _,v in next,{
	{'bp',{'','Use the source'},bp_cmd},
	{'gpsl',{'','Use the source'},bp_gpsl},
	{'iod',{'','Use the source'},iod},
	{'insta',{'','Use the source'},bp_insta},
	{'ch',{'','Use the source'},bp_cheat}
    } do if not commands.commands[v[1]] then
	commands.add_command(table.unpack(v)) end
    end

remote.add_interface('BenchmarkProduction',{
	    insta = bp_insta,
	    cheat = bp_cheat
	}
    )

on('init',function()
    global.trainstop = trainstop
    global.tickstop = tickstop
    global.ioddefault = ioddefault
    global.iodcmd = iodcmd
    for _,t in next,game.forces.player.get_trains() do
	local ts = t.station
	-- s/b `t.stop`^, trains are at particular stops
	if ts then
	    local sr = tsrec(t,ts)
	    sr.discovered_by_startup_scan = true
	    trainstop[t.id] = sr
	    trainstop[sr.ekey] = sr
	    end
	end
    for _,ts in next,game.surfaces[1].find_entities_filtered{
                force=game.forces.player, type='train-stop',
	    } do
	maybe_update_tickstop(ts)
	end
    if not game.forces.benchmark then
	game.create_force('benchmark')
	game.forces.player.set_friend(game.forces.benchmark,true)
	game.forces.benchmark.set_friend(game.forces.player,true)
	end
    end)
on('load',function()
    for k,v in next,global do _G[k]=v end
    if next(tickstop or empty) then
	on('tick',check_tickstop)
	end
    if next(ioddefault or empty) or next(iodcmd or empty) then
	on('marked_for_deconstruction',maybe_iod)
	on('player_deconstructed_area',maybe_finish_iod)
	end
    end)

--   vim: set sts=4 sw=4 tw=71 lbr fo=tcroq com=\:-- :
